"""Implements positional embedding modules for 2D, 3D and time-oriented inputs."""

import abc
import math
import typing

import torch
import torch.fx
from torch import Tensor, nn

from unipercept.ops.embed import fourier_embed_2d
from unipercept.vision.geometry import (
    GridMode,
    GridType,
    generate_coord_grid_like,
    generate_directions_grid,
)

__all__ = ["CoordPos2d", "RadialPos2d", "TrigPos2d", "TrigPos2dTime", "TrigPosTime"]


class _TrigPosBase(nn.Module, metaclass=abc.ABCMeta):
    """Trigonometric positional embedding base class."""

    scale: typing.Final[float | None]
    temperature: typing.Final[float]
    eps: typing.Final[float]
    cache: Tensor | None

    def __init__(
        self,
        dim_input: int,
        *,
        temperature: int = 10000,
        scale: float | None = 2 * math.pi,
        eps: float = 1e-6,
    ) -> None:
        super().__init__()

        self.scale = scale
        self.dim_input = dim_input
        self.temperature = temperature
        self.eps = eps

        self.register_buffer("cache", None, persistent=False)
        self._register_bands(dim_input // 2, dtype=torch.float32, device="cpu")  # type: ignore

    @abc.abstractmethod
    def _generate(self, input: Tensor, mask: Tensor | None = None) -> Tensor: ...

    @abc.abstractmethod
    def _register_bands(self, num_bands: int, **kwargs: typing.Any) -> None: ...  # noqa: ANN401

    @typing.override
    @torch.no_grad()
    def forward(self, input: Tensor, mask: Tensor | None = None) -> Tensor:
        if mask is not None or torch.compiler.is_compiling():
            return self._generate(input, mask)
        if self.cache is None or self.cache.size() != input.size():
            self.cache = self._generate(input, mask)
        return self.cache


class TrigPos2d(_TrigPosBase):
    """Trigonometric positional embedding for 2D inputs."""

    t: Tensor

    @typing.override
    def _register_bands(self, num_bands: int, **kwargs: typing.Any) -> None:
        with torch.no_grad():
            t = torch.arange(num_bands, **kwargs)
            t = self.temperature ** (2 * (t // 2) / num_bands)
        self.register_buffer("t", t, persistent=False)

    @typing.override
    def _generate(self, input: Tensor, mask: Tensor | None = None) -> Tensor:
        """Generate a 2D trigonometric positional embedding."""
        if mask is None:
            H = input.size(-2)
            W = input.size(-1)
            mask = input.new_ones((1, H, W))
        else:
            mask = ~mask
        with torch.autocast(device_type=input.device.type, enabled=False):
            pe = self._embed(mask.float())
        return pe.type_as(input).expand(input.size(0), -1, -1, -1)

    def _embed(self, mask: Tensor) -> Tensor:
        y_embed = mask.cumsum(-2, dtype=torch.float32)
        x_embed = mask.cumsum(-1, dtype=torch.float32)
        t = self.t.to(torch.float32)
        if self.scale is not None:
            y_embed = y_embed / (y_embed[..., -1:, :] + self.eps) * self.scale
            x_embed = x_embed / (x_embed[..., :, -1:] + self.eps) * self.scale
        pos_x = x_embed.unsqueeze(-1) / t
        pos_y = y_embed.unsqueeze(-1) / t
        pos_x = torch.stack(
            (pos_x[..., 0::2].sin(), pos_x[..., 1::2].cos()), dim=4
        ).flatten(3)
        pos_y = torch.stack(
            (pos_y[..., 0::2].sin(), pos_y[..., 1::2].cos()), dim=4
        ).flatten(3)
        return torch.cat((pos_y, pos_x), dim=3).permute(0, 3, 1, 2)


class TrigPos2dTime(_TrigPosBase):
    """Trigonometric positional embedding for 3D/2+1D inputs."""

    t: Tensor
    t_z: Tensor

    @typing.override
    def _register_bands(self, num_bands: int, **kwargs: typing.Any) -> None:
        with torch.no_grad():
            t = torch.arange(num_bands, **kwargs)
            t = self.temperature ** (2 * (t // 2) / num_bands)

            t_z = torch.arange((num_bands * 2), **kwargs)
            t_z = self.temperature ** (2 * (t_z // 2) / (num_bands * 2))

        self.register_buffer("t", t, persistent=False)
        self.register_buffer("t_z", t_z, persistent=False)

    @typing.override
    def _generate(self, input: Tensor, mask: Tensor | None = None) -> Tensor:
        if mask is None:
            T = input.size(-4)
            H = input.size(-2)
            W = input.size(-1)
            mask = input.new_ones((1, T, H, W))
        else:
            mask = ~mask
        with torch.autocast(device_type=input.device.type, enabled=False):
            pe = self._embed(mask.float())
        return pe.type_as(input).expand_as(input)

    def _embed(self, mask: Tensor) -> Tensor:
        """Generate a (2+1)D trigonometric positional embedding."""
        t_embed = mask.cumsum(1)
        y_embed = mask.cumsum(2)
        x_embed = mask.cumsum(3)

        if self.scale is not None:
            t_embed = t_embed / (t_embed[:, -1:, :, :] + self.eps) * self.scale
            y_embed = y_embed / (y_embed[:, :, -1:, :] + self.eps) * self.scale
            x_embed = x_embed / (x_embed[:, :, :, -1:] + self.eps) * self.scale

        pos_x = x_embed[:, :, :, :, None] / self.t
        pos_x = torch.stack(
            (pos_x[:, :, :, :, 0::2].sin(), pos_x[:, :, :, :, 1::2].cos()), dim=5
        ).flatten(4)

        pos_y = y_embed[:, :, :, :, None] / self.t
        pos_y = torch.stack(
            (pos_y[:, :, :, :, 0::2].sin(), pos_y[:, :, :, :, 1::2].cos()), dim=5
        ).flatten(4)

        pos_t = t_embed[:, :, :, :, None] / self.t_z
        pos_t = torch.stack(
            (pos_t[:, :, :, :, 0::2].sin(), pos_t[:, :, :, :, 1::2].cos()), dim=5
        ).flatten(4)

        return (torch.cat((pos_y, pos_x), dim=4) + pos_t).permute(0, 1, 4, 2, 3)


class TrigPosTime(_TrigPosBase):
    """Trigonometric positional embedding for time-oriented inputs."""

    t: Tensor

    @typing.override
    def _register_bands(self, num_bands: int, **kwargs: typing.Any) -> None:
        t = torch.arange(num_bands * 2, requires_grad=False, **kwargs)
        t = self.temperature ** (2 * (t // 2) / num_bands)
        self.register_buffer("t", t, persistent=False)

    @typing.override
    def _generate(self, input: Tensor, mask: Tensor | None = None) -> Tensor:
        if mask is None:
            T = input.size(-2)
            mask = input.new_ones((1, T))
        else:
            mask = ~mask
        with torch.autocast(device_type=input.device.type, enabled=False):
            p = self._embed(mask.float())
        return p.type_as(input).expand_as(input)

    def _embed(
        self,
        mask: Tensor,
    ) -> Tensor:
        """Generate trigonometric positional embedding for time-oriented inputs."""
        t_embed = mask.cumsum(0)
        if self.scale is not None:
            t_embed = t_embed / (t_embed[-1:, :] + self.eps) * self.scale
        pos_t = t_embed[:, :, None] / self.t
        return torch.stack(
            (pos_t[:, :, 0::2].sin(), pos_t[:, :, 1::2].cos()), dim=3
        ).flatten(2)


class CoordPos2d(nn.Module):
    """Embed 2D coordinates."""

    stride: typing.Final[int]
    f_max: typing.Final[float | None]
    cache: Tensor | None

    def __init__(
        self,
        dim_embed: int,
        *,
        stride: int = 1,
        f_max: float | None = None,
        grid_mode: GridType = GridMode.PIXEL_CENTER,
    ) -> None:
        super().__init__()
        self.dim_embed = dim_embed
        self.f_max = f_max
        self.stride = stride
        self.grid_mode = str(grid_mode)

        self.register_buffer("cache", None, persistent=False)

    @typing.override
    @torch.no_grad()
    def forward(self, input: Tensor) -> Tensor:
        if torch.compiler.is_compiling():
            return self._generate(input)
        if self.cache is None or self.cache.size() != input.size():
            self.cache = self._generate(input)
        return self.cache

    def _generate(self, input: Tensor) -> Tensor:
        with torch.autocast(input.device.type, enabled=False):
            pos_matrix = self._generate_grid(input)
            pos_embed = fourier_embed_2d(
                pos_matrix.float(),
                dim_embed=self.dim_embed,
                f_max=self.f_max,
            )
        return pos_embed.type_as(input)

    def _generate_grid(self, input: Tensor) -> Tensor:
        return (
            generate_coord_grid_like(input, mode=self.grid_mode, stride=self.stride)
            .permute(2, 0, 1)
            .unsqueeze(0)
            .repeat(input.size(0), 1, 1, 1)
        )


class RadialPos2d(CoordPos2d):
    """Embed 2D radiance."""

    @typing.override
    def _generate_grid(self, input: Tensor) -> Tensor:
        B, _C, H, W = input.shape

        # Allow scaling using a common stride factor
        H = H * self.stride
        W = W * self.stride

        # Generate homogeneous radial transform as a function of the input size
        r = ((H**2) + (W**2)) ** 0.5
        u = W / 2
        v = H / 2
        transform = input.new_tensor(
            [[[r, 0, u, 0], [0, r, v, 0], [0, 0, 0, 1], [0, 0, 1, 0]]],
        )

        # Generate the directions grid
        dirs = generate_directions_grid(
            transform,
            (H, W),
            dim=1,
            grid_mode=self.grid_mode,
        )
        if self.stride != 1:
            dirs = torch.nn.functional.interpolate(
                dirs,
                scale_factor=1 / self.stride,
                mode="bilinear",
                align_corners=False,
                antialias=False,
            )

        # Repeat along the batch dimension
        return dirs.repeat(B, 1, 1, 1)



"""Commmon utilities for working with strings."""

import re

__all__ = ["convert_word_separator", "to_kebab_case", "to_snake_case"]

GROUP_BY_WORD = re.compile(r"(?<=[a-z\d])(?=[A-Z])|[^a-zA-Z\d\-/]")


def convert_word_separator(string: str, *, sep: str) -> str:
    """Converts a string to a different word separator."""
    return "".join(GROUP_BY_WORD.sub(" ", string).strip().replace(" ", sep))


def to_snake_case(string: str) -> str:
    """Converts a string to snake case."""
    return convert_word_separator(string, sep="_").lower()


def to_kebab_case(string: str) -> str:
    """Converts a string to kebab case."""
    return convert_word_separator(string, sep="-").lower()



"""
Provides utilities for configuring dynamic module namespaces via PEP 562.
"""

import importlib
import importlib.metadata
import importlib.util
import sys
import threading
from collections.abc import Callable, Iterable
from enum import StrEnum, auto
from typing import Any

__all__ = ["ImportMode", "defer_imports"]


class ImportMode(StrEnum):
    """Define the execution behavior for dynamic submodule resolution."""

    DEFERRED = auto()
    LAZY = auto()


def defer_imports(
    name: str,
    modules: Iterable[str],
    /,
    extras: Iterable[str] | None = None,
    mode: ImportMode | str = ImportMode.DEFERRED,
    distribution: str | None = None,
) -> tuple[Callable[[str], Any], Callable[[], list[str]]]:
    """
    Create getattr and dir functions for dynamic module attribute resolution.

    Parameters
    ----------
    name : str
        The absolute name of the parent module.
    modules : Iterable[str]
        Submodules managed by the namespace hooks.
    extras : Iterable[str], optional
        Static attributes to include in the module's directory listing.
    mode : ImportMode | str, default ImportMode.DEFERRED
        The execution model applied upon attribute resolution. DEFERRED executes
        the target module immediately upon access. LAZY defers execution until
        a sub-attribute is accessed.
    distribution : str, optional
        The name of the distribution this module belongs to. If defined, a
        __version__ attribute will be added to the module.

    Returns
    -------
    __getattr__ : Callable[[str], Any]
        A function conforming to the __getattr__ specification.
    __dir__ : Callable[[], list[str]]
        A function conforming to the __dir__ specification.
    """
    module_set = frozenset(modules)
    directory_set = set(module_set) | set(extras or [])
    if distribution is not None:
        directory_set.add("__version__")
    directory = tuple(sorted(directory_set))
    lock = threading.Lock()

    # Fail early on invalid configuration rather than at attribute access time
    try:
        active_mode = ImportMode(mode)
    except ValueError:
        valid_modes = [m.value for m in ImportMode]
        msg = f"Invalid mode {mode!r}. Expected one of: {valid_modes}"
        raise ValueError(msg)  # noqa: B904

    def __getattr__(attr_name: str) -> Any:
        # print(f"DEBUG: __getattr__({name=}, {attr_name=})")
        if attr_name == "__version__" and distribution is not None:
            try:
                return importlib.metadata.version(distribution)
            except importlib.metadata.PackageNotFoundError:
                return "unknown"

        if attr_name in module_set:
            fullname = f"{name}.{attr_name}"
            # print(f"DEBUG: importing {fullname=}")
            module = importlib.import_module(fullname)
            setattr(sys.modules[name], attr_name, module)
            return module

        msg = f"Module {name!r} has no attribute {attr_name!r}"
        raise AttributeError(msg)

    def __dir__() -> list[str]:
        return list(directory)

    return __getattr__, __dir__



from __future__ import annotations

import enum as E

import torch
import torch.signal
import torchvision.transforms.v2.functional as tvfn
from torch import Tensor, nn

from unipercept.utils.signal import get_gaussian_2d

__all__ = ["masks_to_boxes", "masks_to_centers"]


def masks_to_centers(masks: Tensor, stride: int = 1, use_vmap: bool = False) -> Tensor:
    """
    Converts a mask to a center point.

    Parameters
    ----------
    masks
        Mask tensor with shape (N, H, W)
    stride
        Scaling factor of the mask.

    Returns
    -------
        Center point tensor (N, (X,Y)), where X in [0,W] and Y in [0,H]
    """
    if masks.numel() == 0:
        return torch.zeros((0, 2), device=masks.device, dtype=masks.dtype)

    masks = masks.permute(0, 2, 1)  # Ensure output is XY not YX
    axes = torch.stack(_get_index_axes_like(masks, stride=stride), dim=-1)

    if use_vmap:
        return torch.vmap(_get_mass_center, (None, -1), (1))(masks, axes)
    cmap = []
    for i in range(axes.shape[-1]):
        cmap.append(_get_mass_center(masks, axes[..., i]))
    return torch.stack(cmap, dim=-1)


@torch.no_grad()
def masks_to_boxes(
    masks: Tensor,
    stride: int = 1,
    filter_size: int | tuple[int, int] | None = None,
    threshold: float = 0.5,
) -> Tensor:
    """
    Convert masks to bounding boxes.

    Parameters
    ----------
    masks
        Mask tensor with shape (N, H, W)
    stride
        Scaling factor of the mask.
    filter_size
        Size of the kernel used to blur the mask. By default the size is 15.
    filter_threshold
        Threshold for the filter. By default the threshold is 0.33.

    Returns
    -------
        Bounding box tensor (N, (X1,Y1,X2,Y2)), where X in [0,W] and Y in [0,H]
    """
    if masks.numel() == 0:
        return torch.zeros((0, 4), device=masks.device)
    if masks.ndim != 3:
        msg = f"Expected masks to have 3 dimensions, got {masks.ndim}"
        raise ValueError(msg)

    if filter_size is not None and threshold > 0.0:
        masks_blur = masks.float()
        masks_blur = tvfn.gaussian_blur_image(masks_blur, filter_size)
        masks_blur = masks >= threshold
        masks_blur_valid = masks_blur.int().sum(dim=(-1, -2)) > 0
        masks[masks_blur_valid] = masks_blur[masks_blur_valid]

    if masks.dtype != torch.bool:
        masks = masks >= threshold

    axes = _get_index_axes_like(masks, stride=stride)
    return _get_bounding_box_batched(axes, masks)


@torch.no_grad()
def _get_index_axes_like(t: Tensor, *, stride: int) -> Tensor:
    """Returns a grid of indices with shape (H, W, 2)."""
    h, w = t.shape[-2:]
    with t.device:
        y = torch.arange(0, h * stride, stride, dtype=torch.float) + stride // 2
        x = torch.arange(0, w * stride, stride, dtype=torch.float) + stride // 2
    # g = torch.stack(torch.meshgrid(y, x, indexing="ij"), dim=-1)
    return torch.meshgrid(y, x, indexing="ij")


def _get_mass_center(masks: Tensor, index_axes: Tensor) -> Tensor:
    sum_pixels = (masks * index_axes).sum(dim=(-1, -2))
    num_pixels = masks.sum(dim=(-1, -2)).clamp(min=1.0)
    return sum_pixels / num_pixels

    # nanmean not yet implemented in Inductor?
    # return (m / m * ax).nanmean(dim=(-2, -1))


def _get_bounding_box_batched(index_axes: Tensor, masks: Tensor):
    y, x = index_axes

    x_mask = masks * x.unsqueeze(0)
    x_max = x_mask.flatten(1).max(-1)[0]
    x_min = x_mask.masked_fill(~(masks.bool()), 1e8).flatten(1).min(-1)[0]

    y_mask = masks * y.unsqueeze(0)
    y_max = y_mask.flatten(1).max(-1)[0]
    y_min = y_mask.masked_fill(~(masks.bool()), 1e8).flatten(1).min(-1)[0]

    return torch.stack([x_min, y_min, x_max, y_max], 1)


class BlurSizeMethod(E.StrEnum):
    SHORT = E.auto()
    LONG = E.auto()
    MEAN = E.auto()


def blur_masks(
    masks: Tensor,
    alpha: float = 0.15,
    eps: float = 0.35,
    k_factor: float = 1.0,
    k_method: BlurSizeMethod = BlurSizeMethod.SHORT,
) -> Tensor:
    if masks.numel() == 0:
        return masks

    dtype = masks.dtype
    ndim = masks.ndim
    if ndim == 2:
        masks = masks.unsqueeze(dim=0)
    shape = masks.shape

    ltbr = masks_to_boxes(masks)
    w = ltbr[:, 2] - ltbr[:, 0]
    h = ltbr[:, 3] - ltbr[:, 1]

    match k_method:
        case BlurSizeMethod.SHORT:
            k_sizes = torch.min(w, h)
        case BlurSizeMethod.LONG:
            k_sizes = torch.max(w, h)
        case BlurSizeMethod.MEAN:
            k_sizes = (w + h) / 2
        case _:
            msg = f"Invalid blur size method: {k_method}"
            raise ValueError(msg)

    k_sizes = (k_sizes * k_factor).int()
    k_sizes = k_sizes // 2 * 2 + 1
    k_sizes = k_sizes.clamp(3)

    if ndim in {2, 3}:
        masks = masks.unsqueeze(dim=0)
    elif ndim > 4:
        masks = masks.reshape((-1, *shape[-3:]))

    fp = torch.is_floating_point(masks)

    k_size_wmax = k_sizes.max().item()
    k_size_hmax = k_size_wmax
    ks_list: list[Tensor] = []
    for k_size in k_sizes.unbind(0):
        k_item = k_size.item()
        sigma = alpha * k_item + eps

        k = get_gaussian_2d(
            [k_item, k_item],
            [sigma, sigma],
            dtype if fp else torch.float32,
            device=masks.device,
        )

        k_size_wdiff = k_size_wmax - k_item
        k_size_hdiff = k_size_hmax - k_item
        k = torch.nn.functional.pad(
            k,
            [
                k_size_wdiff // 2,
                k_size_wdiff // 2,
                k_size_hdiff // 2,
                k_size_hdiff // 2,
            ],
        )
        ks_list.append(k)

    ks = torch.stack(ks_list, dim=0).unsqueeze(1)

    assert ks.ndim == 4, ks.shape  # (N, 1, K, K)

    output = masks if fp else masks.to(dtype=torch.float32)

    # padding = (left, right, top, bottom)
    padding = [
        k_size_wmax // 2,
        k_size_wmax // 2,
        k_size_hmax // 2,
        k_size_hmax // 2,
    ]
    output = nn.functional.pad(output, padding, mode="reflect")
    output = nn.functional.conv2d(output, ks, groups=shape[-3])

    if ndim == 2:
        output = output.squeeze(dim=0).squeeze(dim=0)
    elif ndim == 3:
        output = output.squeeze(dim=0)
    elif ndim > 4:
        output = output.reshape(shape)

    if not fp:
        output = output.round_().to(dtype=dtype)

    return output


if __name__ == "__main__":
    # Example usage, this draws a bounding box around the masks and a circle at the
    # center of the mask.
    import matplotlib.patches as pat
    import matplotlib.pyplot as plt
    import torch

    import unipercept as up

    mask1 = torch.zeros(128, 128)
    mask1[32:96, 32:96] = 1

    mask2 = torch.zeros(128, 128)
    mask2[64:76, 64:120] = 1

    masks = torch.stack([torch.zeros_like(mask1), mask1, mask2])
    boxes = up.utils.mask.masks_to_boxes(masks)
    centers = up.utils.mask.masks_to_centers(masks)

    fig, ax = plt.subplots()
    ax.imshow(masks.argmax(0))

    for xyxy in boxes.unbind(0):
        x1, y1, x2, y2 = map(float, xyxy.tolist())
        w = x2 - x1
        h = y2 - y1

        ax.add_patch(pat.Rectangle((x1, y1), w, h, fill=False, edgecolor="red"))

    for xy in centers:
        x, y = map(int, xy.tolist())
        ax.add_patch(pat.Circle((x, y), radius=2, fill=True, color="red"))



"""Implements interpolation layers."""

from __future__ import annotations

import torch
import typing_extensions as TX
from torch import nn


class Interpolate2d(nn.Module):
    r"""Wraps :func:`torch.nn.functional.interpolate` as a module."""

    name: torch.jit.Final[str]
    size: torch.jit.Final[int | tuple[int, int] | None]
    scale_factor: torch.jit.Final[float | tuple[float, float] | None]
    mode: torch.jit.Final[str]
    align_corners: torch.jit.Final[bool | None]

    def __init__(
        self,
        *,
        size: int | tuple[int, int] | None = None,
        scale_factor: float | tuple[float, float] | None = None,
        mode: str = "nearest",
        align_corners: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.name = type(self).__name__
        self.size = size
        if isinstance(scale_factor, tuple):
            assert len(scale_factor) == 2, "scale_factor must be a 2-tuple"
            self.scale_factor = (float(scale_factor[0]), float(scale_factor[1]))
        else:
            self.scale_factor = float(scale_factor) if scale_factor else None
        self.mode = mode
        self.align_corners = None if mode == "nearest" else align_corners

    @TX.override
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return nn.functional.interpolate(
            input,
            self.size,
            self.scale_factor,
            self.mode,
            self.align_corners,
            recompute_scale_factor=False,
        )



from __future__ import annotations

import typing as T

import torch
from torch import nn

EPS_FLOAT: T.Final[float] = torch.finfo(torch.float).resolution
EPS_HALF: T.Final[float] = torch.finfo(torch.half).resolution
EPS_BF16: T.Final[float] = torch.finfo(torch.bfloat16).resolution


class ScaledLossMixin(nn.Module):
    __constants__ = ("scale",)

    def __init__(self, *, scale: float = 1.0, **kwargs) -> None:
        super().__init__(**kwargs)

        self.scale = scale


class StableLossMixin(nn.Module):
    __constants__ = ("eps",)

    def __init__(self, *, eps: float = EPS_FLOAT, **kwargs) -> None:
        super().__init__(**kwargs)

        self.eps = eps
